from aigyminsper.search.SearchAlgorithms import BuscaLargura
from aigyminsper.search.SearchAlgorithms import BuscaProfundidade
from aigyminsper.search.SearchAlgorithms import BuscaProfundidadeIterativa
from .SumOne import SumOne

def test_largura():
    state = SumOne(1, '', 11)
    algorithm = BuscaLargura()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2 "

def test_largura2():
    state = SumOne(1, '', 23)
    algorithm = BuscaLargura()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2 "

def test_largura3():
    state = SumOne(1, '', 5)
    algorithm = BuscaLargura()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2 "

def test_profundidade():
    state = SumOne(1, '', 10)
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 10)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +1 "

def test_BPI():
    state = SumOne(1, '', 10)
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +1 "



from aigyminsper.search.CSPAlgorithms import SubidaMontanha, SubidaMontanhaEstocastico
from .Equacao import Equacao
import pytest

"""
Testes com subida da montanha
"""

DELTA = 0.1


@pytest.mark.parametrize(
    "n, expected",
    [(0, 2.18), (2, 2.18), (4, 2.18), (7, 8.07), (10, 8.07), (11, 14.26), (15, 14.26)],
)
def test_equacao_subida_montanha(n, expected):
    state = Equacao("", n)
    algorithm = SubidaMontanha()
    result = algorithm.search(state)
    assert result.number == pytest.approx(expected, DELTA)


"""
Testes com subida da montanha estocástico
"""


@pytest.mark.parametrize(
    "n, expected",
    [
        (0, 14.26),
        (2, 14.26),
        (4, 14.26),
        (7, 14.26),
        (10, 14.26),
        (11, 14.26),
        (15, 14.26),
    ],
)
def test_equacao_subida_montanha_estocastico(n, expected):
    state = Equacao("", n)
    algorithm = SubidaMontanhaEstocastico()
    result = algorithm.search(state)
    assert result.number == pytest.approx(expected, DELTA)



#
# Considerando as implementações: 
#
# - AspiradorPo.py
# - SumOne.py
# - poi.py
#
# Testar o algoritmo: 
#
# 2) Depth-first search (BuscaProfundidade)
#
# Com as seguintes configurações de poda: 
#
# - without pruning
# - father-son pruning
# - general pruning
#
#

from aigyminsper.search.SearchAlgorithms import BuscaProfundidade
from .SumOne import SumOne
from .poi import Poi
from .AspiradorPo import AspiradorPo

# ---------------------------------- SumOne ---------------------------------- #

def test_bp_sumone_without_pruning():
    state = SumOne(1, '', 20)
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 20, pruning='without')
    assert result.show_path() == ' ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +1 '
    assert result.g == 10

def test_bp_sumone_father_son_pruning():
    state = SumOne(1, '', 20)
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 20, pruning='father-son')
    assert result.show_path() == ' ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +1 '
    assert result.g == 10

def test_bp_sumone_general_pruning():
    state = SumOne(1, '', 20)
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 20, pruning='general')
    assert result.show_path() == ' ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +1 '
    assert result.g == 10


# ---------------------------------- POI ---------------------------------- #

def test_bp_poi_without_pruning():
    state = Poi('', 'A', 'A', 'E')
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 5, pruning='without')
    assert result.show_path() == ' ; ir p/ c ; ir p/ b ; ir p/ a ; ir p/ c ; ir p/ e'
    assert result.g == 27

def test_bp_poi_father_son_pruning():
    state = Poi('', 'A', 'A', 'E')
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 5, pruning='father-son')
    assert result.show_path() == ' ; ir p/ c ; ir p/ b ; ir p/ a ; ir p/ c ; ir p/ e'
    assert result.g == 27

def test_bp_poi_general_pruning():
    state = Poi('', 'A', 'A', 'E')
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 5, pruning='general')
    assert result.show_path() == ' ; ir p/ c ; ir p/ b ; ir p/ d ; ir p/ e'
    assert result.g == 15


# ---------------------------------- Aspirador de Pó ---------------------------------- #

def test_bp_aspiradorpo_without_pruning():
    state = AspiradorPo('','ESQ','SUJO','SUJO')
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 5, pruning='without')
    assert result.show_path() == ' ; limpar ; limpar ; dir ; limpar ; esq'
    assert result.g == 5

def test_bp_aspiradorpo_father_son_pruning():
    state = AspiradorPo('','ESQ','SUJO','SUJO')
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 10, pruning='father-son')
    assert result.show_path() == ' ; limpar ; dir ; limpar ; esq'
    assert result.g == 4

def test_bp_aspiradorpo_general_pruning():
    state = AspiradorPo('', 'ESQ', 'SUJO', 'SUJO')
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 5, pruning='general')
    assert result.show_path() == ' ; limpar ; dir ; limpar ; esq'

def test_bp_aspiradorpo_without_pruning_10():
    state = AspiradorPo('', 'ESQ', 'SUJO', 'SUJO')
    algorithm = BuscaProfundidade()
    result = algorithm.search(state, 10, pruning='without')
    assert result.show_path() == ' ; limpar ; limpar ; limpar ; limpar ; limpar ; limpar ; limpar ; dir ; limpar ; esq'


from aigyminsper.search.SearchAlgorithms import BuscaProfundidadeIterativa
from .SumOne import SumOne
from .poi import Poi
from .AspiradorPo import AspiradorPo

def test_sumone_without_pruning():
    state = SumOne(3, '', 6)
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +1 "

def test_sumone_father_son_pruning():
    state = SumOne(3, '', 6)
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state, pruning='father-son')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +1 "

def test_sumone_general_pruning():
    state = SumOne(3, '', 6)
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state, pruning='general')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +1 "

def test_poi_without_pruning():
    state = Poi('','A','A','D')
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state)
    assert result.show_path() == " ; ir p/ b ; ir p/ d"

def test_poi_father_son_pruning():
    state = Poi('','A','A','D')
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state, pruning='father-son')
    assert result.show_path() == " ; ir p/ b ; ir p/ d"

def test_poi_general_pruning():
    state = Poi('','A','A','D')
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state, pruning='general')
    assert result.show_path() == " ; ir p/ b ; ir p/ d"

def test_aspiradorpo_without_pruning():
    state = AspiradorPo('','ESQ','LIMPO','SUJO')
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state)
    assert result.show_path() == " ; dir ; limpar ; esq"

def test_aspiradorpo_father_son_pruning():
    state = AspiradorPo('','ESQ','LIMPO','SUJO')
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state, pruning='father-son')
    assert result.show_path() == " ; dir ; limpar ; esq"
    
def test_aspiradorpo_general_pruning():
    state = AspiradorPo('','ESQ','LIMPO','SUJO')
    algorithm = BuscaProfundidadeIterativa()
    result = algorithm.search(state, pruning='general')
    assert result.show_path() == " ; dir ; limpar ; esq"



from abc import ABC, abstractmethod

class Node:
    """
    It is the most basic data structures necessary to implement search algorithms.
    """

    def __init__(self,state,father_node):
        """
        - state: the state represented by the node
        - father_node: the father node of the current node
        """

        self.state = state
        self.father_node = father_node
        if self.father_node == None:
            self.depth = 0
            self.g = 0
        else:
            self.depth = father_node.depth + 1
            self.g = state.cost() + self.father_node.g

    def show_path(self):
        """
        Return the path from the root node to the current node
        """
        if self.father_node != None:
            return self.father_node.show_path()  + " ; " + self.state.operator 
        else:
            return self.state.operator
    
    def h(self):
        """
        Return the heuristic value of the current node
        """
        return self.state.h()

    def f(self):
        """
        Return the evaluation function value of the current node: f(n) = g(n) + h(n)
        """
        return self.g + self.h()
     
    
class State(ABC):
    """
    This class represents a state in a search problem.
    This is an abstract class. This class defines the following abstract methods:
    - successors: returns a list of successors of the current state
    - is_goal: returns True if the current state is a goal state
    - description: returns a string with a brief description of the problem
    - cost: returns the cost of the current state
    - env: returns the description of the environment of the current state

    This class also defines the following non-abstract methods:
    - print: returns a string with the operator that generated the current state
    """

    @abstractmethod
    def successors(self):
        """
        Return a list of successors of the current state
        """
        pass
    
    @abstractmethod
    def is_goal(self):
        """
        It returns True if the current state is a goal state
        """
        pass
    
    @abstractmethod
    def description(self):
        """
        Return a string with a brief description of the problem
        """
        pass
    
    @abstractmethod
    def cost(self):
        """
        Return the cost of the current state
        """
        pass

    def print(self):
        """
        Return a string with the operator that generated the current state
        """
        return str(self.operator)

    @abstractmethod
    def env(self):
        """
        Return the description of the environment of the current state
        """
        pass


from collections import deque
from aigyminsper.search.Graph import Node

def sortFunction(val):
    """
    Function to sort the list by g(), h() or f()  
    """
    return val[1]

class SearchAlgorithm:
    """
    This class implements an interface for search algorithms.
    This class should not be instantiated.

    This class is used by the following implementations:
        - Breadth-first search (BuscaLargura)
        - Depth-first search (BuscaProfundidade)
        - Iterative deepening search (BPI)
        - Uniform cost search (CustoUniforme)
        - Greddy search algorithm (BuscaGananciosa)
        - A* search algorithm (AEstrela)
    """

    def search(self, initialState, pruning='without', trace=False):
        """
        This method implements a search algorithm.

        Parameters:
            initialState: the initial state of the search.
            pruning: a string that defines the pruning option. The pruning options are: without, father-son and general.
            trace: a boolean that defines if the trace if printed or not.
        """
        pass

    def print_trace(self, node) -> None:
        """
        This method prints the trace of the search.

        Parameters:
            node: the node that is the solution of the search.
        """
        print(f"Path (State {node.state.env()}): {node.show_path()} -- Cost: {node.g}")


class BuscaLargura (SearchAlgorithm):
    """
    This class implements the Breadth-first search algorithm.    
    """

    def search (self, initialState, pruning='without', trace=False): 
        # Define valid pruning options
        valid_pruning_options = ['without', 'father-son', 'general']
        if pruning not in valid_pruning_options:
            raise ValueError(f"Invalid pruning option: {pruning}. Valid options are {valid_pruning_options}")

        # Set to keep track of the visited nodes
        states = set()
        #Creating a Queue
        open = deque()
        open.append(Node(initialState, None))
        while (len(open) > 0):
            n = open.popleft()
            if trace: self.print_trace(n)

            if (n.state.is_goal()):
                return n
            for i in n.state.successors():
                new_n = Node(i,n)
                # without pruning
                if pruning == "without":
                    open.append(new_n)
                # father-son pruning
                elif pruning == "father-son" and (new_n.state.env() != n.state.env()):
                    open.append(new_n)
                # general pruning
                elif pruning == "general" and (new_n.state.env() not in states):
                    open.append(new_n)
                    states.add(new_n.state.env())
        return None


class BuscaProfundidade (SearchAlgorithm):
    """
    This class implements the Depth-first search (limited)    
    """

    def search (self, initialState, m, pruning='without', trace=False): 
        # Define valid pruning options
        valid_pruning_options = ['without', 'father-son', 'general']
        if pruning not in valid_pruning_options:
            raise ValueError(f"Invalid pruning option: {pruning}. Valid options are {valid_pruning_options}")

        # Set to keep track of the visited nodes
        states = set()
        #Using list as stack
        open = []
        open.append(Node(initialState, None))
        while (len(open) > 0):
            n = open.pop()
            if trace: self.print_trace(n)

            if (n.state.is_goal()):
                return n
            if (n.depth < m):
                for i in n.state.successors():
                    new_n = Node(i,n)
                    # without pruning
                    if pruning == "without":
                        open.append(new_n)
                    # father-son pruning
                    elif pruning == "father-son" and (new_n.state.env() != n.state.env()):
                        open.append(new_n)
                    # general pruning
                    elif pruning == "general" and (new_n.state.env() not in states):
                        open.append(new_n)
                        states.add(new_n.state.env())
        return None

class BuscaProfundidadeIterativa (SearchAlgorithm):
    """
    This class implements Iterative Deepening Depth-first search
    """

    def search (self, initialState, pruning='without', trace=False): 
        n = 1
        algorithm = BuscaProfundidade()
        while True:
            result = algorithm.search(initialState, n, pruning, trace)
            if (result != None):
                return result
            n = n+1


class BuscaCustoUniforme (SearchAlgorithm):
    """
    This class implements a Uniform cost search algorithm
    """

    def search (self, initialState, pruning='without', trace=False):
        # Define valid pruning options
        valid_pruning_options = ['without', 'father-son', 'general']
        if pruning not in valid_pruning_options:
            raise ValueError(f"Invalid pruning option: {pruning}. Valid options are {valid_pruning_options}")

        # Set to keep track of the visited nodes
        states = set()
        open = []
        new_n = Node(initialState, None)
        open.append((new_n, new_n.g))
        while (len(open) > 0):
            #list sorted by g()
            open.sort(key = sortFunction, reverse = True)
            n = open.pop()[0]
            if trace: self.print_trace(n)

            if (n.state.is_goal()):
                return n
            for i in n.state.successors():
                new_n = Node(i,n)
                # without pruning
                if pruning == "without":
                    open.append((new_n, new_n.g))
                # father-son pruning
                elif pruning == "father-son" and (new_n.state.env() != n.state.env()):
                    open.append((new_n, new_n.g))
                # general pruning
                elif pruning == "general" and (new_n.state.env() not in states):
                    open.append((new_n, new_n.g))
                    states.add(new_n.state.env())
        return None
    

class BuscaGananciosa (SearchAlgorithm):
    """
    This class implements a Greddy search algorithm
    """

    def search (self, initialState, pruning='without', trace=False):
        # Define valid pruning options
        valid_pruning_options = ['without', 'father-son', 'general']
        if pruning not in valid_pruning_options:
            raise ValueError(f"Invalid pruning option: {pruning}. Valid options are {valid_pruning_options}")

        # Set to keep track of the visited nodes
        states = set()
        open = []
        new_n = Node(initialState, None)
        open.append((new_n, new_n.h()))
        while (len(open) > 0):
            #list sorted by h()
            open.sort(key = sortFunction, reverse = True)
            n = open.pop()[0]
            if trace: self.print_trace(n)

            if (n.state.is_goal()):
                return n
            for i in n.state.successors():
                new_n = Node(i,n)
                # without pruning
                if pruning == "without":
                    open.append((new_n, new_n.h()))
                # father-son pruning
                elif pruning == "father-son" and (new_n.state.env() != n.state.env()):
                    open.append((new_n, new_n.h()))
                # general pruning
                elif pruning == "general" and (new_n.state.env() not in states):
                    open.append((new_n, new_n.h()))
                    states.add(new_n.state.env())
        return None


class AEstrela (SearchAlgorithm):
    """
    This class implements a A* search algorithm
    """

    def search (self, initialState, pruning='without', trace=False):
        # Define valid pruning options
        valid_pruning_options = ['without', 'father-son', 'general']
        if pruning not in valid_pruning_options:
            raise ValueError(f"Invalid pruning option: {pruning}. Valid options are {valid_pruning_options}")

        # Set to keep track of the visited nodes
        states = set()
        open = []
        new_n = Node(initialState, None)
        open.append((new_n, new_n.f()))
        while (len(open) > 0):
            #list sorted by f()
            open.sort(key = sortFunction, reverse = True)
            n = open.pop()[0]
            if trace: self.print_trace(n)
            
            if (n.state.is_goal()):
                return n
            # iterate trought all successors
            for i in n.state.successors():

                new_n = Node(i,n)
                # without pruning
                if pruning == "without":
                    open.append((new_n,new_n.f()))
                # father-son pruning
                elif pruning == "father-son" and (new_n.state.env() != n.state.env()):
                    open.append((new_n,new_n.f()))
                # general pruning
                elif pruning == "general" and (new_n.state.env() not in states):
                    open.append((new_n,new_n.f()))
                    # nao eh adiciona o estado ao vetor.
                    # eh adicionado o conteudo
                    states.add(new_n.state.env())
        return None



#
# Considerando as implementações: 
#
# - AspiradorPo.py
# - SumOne.py
# - poi.py
#
# Testar o algoritmo: 
#
# 1) Breadth-first search (BuscaLargura)
#
# Com as seguintes configurações de poda: 
#
# - without pruning
# - father-son pruning
# - general pruning
#
#

from aigyminsper.search.SearchAlgorithms import BuscaLargura
from .SumOne import SumOne
from .poi import Poi

def test_sumone_without_pruning():
    state = SumOne(1, '', 11)
    algorithm = BuscaLargura()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2 "

def test_sumone_father_pruning():
    state = SumOne(1, '', 11)
    algorithm = BuscaLargura()
    result = algorithm.search(state, pruning='father-son')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2 "

def test_sumone_general():
    state = SumOne(1, '', 11)
    algorithm = BuscaLargura()
    result = algorithm.search(state, pruning='general')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2 "

def test_poi_BL():
    state = Poi('','0','A','E')
    alg = BuscaLargura()
    result = alg.search(state)
    assert result.show_path() == ' ; ir p/ c ; ir p/ e'

def test_poi_BL_custo():
    state = Poi('','0','A','E')
    alg = BuscaLargura()
    result = alg.search(state)
    assert result.g == 15

def test_poi_BL_custo_0():
    state = Poi('','0','A','A')
    alg = BuscaLargura()
    result = alg.search(state)
    assert result.g == 0



"""
Exposes the main classes of the search module:

- State: the class that represents a state of the search.
- SearchAlgorithm: the class that represents a search algorithm.
- BuscaLargura: the class that implements the Breadth-first search algorithm.
- BuscaProfundidade: the class that implements the Depth-first search algorithm.
- BuscaProfundidadeIterativa: the class that implements the Iterative Depth-first search algorithm.
- BuscaCustoUniforme: the class that implements the Uniform cost search algorithm.
- BuscaGananciosa: the class that implements the Greedy search algorithm.
- AEstrela: the class that implements the A* search algorithm.
- SubidaMontanha: the class that implements the Hill Climbing search algorithm.
- SubidaMontanhaEstocastico: the class that implements the Stochastic Hill Climbing search algorithm.
"""

from .Graph import State
from .SearchAlgorithms import (
    SearchAlgorithm,
    BuscaLargura,
    BuscaProfundidade,
    BuscaProfundidadeIterativa,
    BuscaCustoUniforme,
    BuscaGananciosa,
    AEstrela,
)
from .CSPAlgorithms import SubidaMontanha, SubidaMontanhaEstocastico

__all__ = [
    "State",
    "SearchAlgorithm",
    "BuscaLargura",
    "BuscaProfundidade",
    "BuscaProfundidadeIterativa",
    "BuscaCustoUniforme",
    "BuscaGananciosa",
    "AEstrela",
    "SubidaMontanha",
    "SubidaMontanhaEstocastico",
]



#
# Considerando as implementações: 
#
# - AspiradorPo.py
# - SumOne.py
# - poi.py
#
# Testar o algoritmo: 
#
# 4) Uniform cost search (CustoUniforme)
#
# Com as seguintes configurações de poda: 
#
# - without pruning
# - father-son pruning
# - general pruning
#
#

from aigyminsper.search.SearchAlgorithms import BuscaCustoUniforme
from .SumOne import SumOne
from .poi import Poi
from .AspiradorPo import AspiradorPo

def test_bcu_sumone_without_pruning():
    state = SumOne(1, '', 23)
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2 "
    assert result.g == 11

def test_bcu_sumone_father_son_pruning():
    state = SumOne(1, '', 23)
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state, pruning='father-son')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2 "
    assert result.g == 11

def test_bcu_sumone_general_pruning():
    state = SumOne(1, '', 23)
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state, pruning='general')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2  ; +2 "
    assert result.g == 11

def test_bcu_aspirador_without_pruning():
    state = AspiradorPo('','ESQ','SUJO','SUJO')
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; dir ; limpar ; esq ; limpar"
    assert result.g == 4

def test_bcu_aspirador_father_son_pruning():
    state = AspiradorPo('','ESQ','SUJO','SUJO')
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state, pruning='father-son')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; dir ; limpar ; esq ; limpar"
    assert result.g == 4

def test_bcu_aspirador_general_pruning():
    state = AspiradorPo('','ESQ','SUJO','SUJO')
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state, pruning='general')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; limpar ; dir ; limpar ; esq"
    assert result.g == 4

def test_bcu_poi_without_pruning():
    state = Poi('',"A","A","G")
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state)
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; ir p/ b ; ir p/ d ; ir p/ e ; ir p/ f ; ir p/ g"
    assert result.g == 7

def test_bcu_poi_father_son_pruning():
    state = Poi('',"A","A","G")
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state, pruning='father-son')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; ir p/ b ; ir p/ d ; ir p/ e ; ir p/ f ; ir p/ g"
    assert result.g == 7

def test_bcu_poi_general_pruning():
    state = Poi('',"A","A","G")
    algorithm = BuscaCustoUniforme()
    result = algorithm.search(state, pruning='general')
    print(f'\nEstado inicial = {state.env()}')
    print(f'Solução = {result.show_path()}')
    assert result.show_path() == " ; ir p/ b ; ir p/ d ; ir p/ e ; ir p/ f ; ir p/ g"
    assert result.g == 7


#
# Considerando as implementações: 
#
# - AspiradorPo.py
# - SumOne.py
# - poi.py
#
# Testar o algoritmo: 
#
# 5) Greddy search algorithm (BuscaGananciosa)
#
# Com as seguintes configurações de poda: 
#
# - without pruning
# - father-son pruning
# - general pruning
#
#


from setuptools import setup

with open('./docs/index.md') as f:
    long_description = f.read()

setup(
    name='aigyminsper',
    version='0.2.0',
    packages=['aigyminsper','aigyminsper.search'],
    install_requires=[],
    description='A libray that helps you to learn Artificial Intelligence.',
    author='Fabricio Barth',
    author_email='fabriciojb@insper.edu.br',
    long_description=long_description,
    long_description_content_type='text/markdown',
)



"""
This module implements the following search algorithms:

- hill-climing search algorithm
- stochastic hill-climbing search algorithm
- simulated annealing search (algoritmo da têmpera simulada) TODO
- local beam search (busca em feixe local) TODO

"""

from aigyminsper.search.SearchAlgorithms import SearchAlgorithm

def sortFunction(val):
    """
    Function to sort the list by g(), h() or f()  
    """
    return val[1]


class SubidaMontanha (SearchAlgorithm):
    """
    This class implements the hill-climbing search algorithm.
    """

    def best(self, successors):
        best_state = successors[0]
        for i in successors:
            if i.h() < best_state.h():
                best_state = i
        return best_state

    def search (self, initialState, trace= False):
        atual = initialState
        while True:
            if trace: print(f'Estado = {atual.env()}') 
            prox = self.best(atual.successors())
            if prox.h() >= atual.h():
                return atual
            atual = prox


class SubidaMontanhaEstocastico (SearchAlgorithm):
    """
    This class implements the stochastic hill-climbing search algorithm.
    """

    def best(self, successors):
        best_state = successors[0]
        for i in successors:
            if i.h() < best_state.h():
                best_state = i
        return best_state

    def search (self, initialState, trace= False):
        atual = initialState
        while True:
            if trace: print(f'Estado = {atual.env()}') 
            prox = self.best(atual.successors())
            if prox.h() >= atual.h():
                if atual.is_goal():
                    return atual
                else: 
                    atual.randomState()
            else:
                atual = prox


